from collections import Counter
import polars as pl
import intervaltree as it

from typing import Any, Callable, NamedTuple

from .constants import (
    DEF_BP_MERGE_BLKS,
    DEF_BP_MERGE_UNITS,
    DEF_MERGE_RBLACKLIST,
    DEF_MIN_ARR_HOR_UNITS,
    DEF_MIN_BLK_HOR_UNITS,
    DEF_MIN_ARR_PROP,
    DEF_MIN_ARR_LEN,
    DEF_OUTPUT_BED_COLS,
    DEF_OUTPUT_BED_COLS_STRAND,
)
from ..common import merge_itvs


class MergeHORData(NamedTuple):
    name: str
    hor_count: int
    hor_len: int


def bed_to_itree(df: pl.DataFrame, fn_mdata: Callable[[int, int], Any] | None = None):
    return it.IntervalTree(
        it.Interval(st, end, fn_mdata(st, end) if fn_mdata else None)
        for st, end in df.select("chrom_st", "chrom_end").iter_rows()
    )


def merge_hor_unit_itvs(i1: it.Interval, i2: it.Interval) -> it.Interval:
    i1_data: MergeHORData = i1.data
    i2_data: MergeHORData = i2.data
    # Keep count of number of intervals merged and merged length.
    new_data = MergeHORData(
        i1_data.name, i1_data.hor_count + 1, i1_data.hor_len + i2_data.hor_len
    )
    return it.Interval(i1.begin, i2.end, new_data)


def group_by_dst(df: pl.DataFrame, dst: int, group_name: str) -> pl.DataFrame:
    try:
        df = df.drop("index")
    except pl.exceptions.ColumnNotFoundError:
        pass
    return (
        df.with_columns(
            # c1  st1 (end1)
            # c1 (st2) end2
            dst_behind=(pl.col("chrom_st") - pl.col("chrom_end").shift(1)).fill_null(0),
            dst_ahead=(pl.col("chrom_st").shift(-1) - pl.col("chrom_end")).fill_null(0),
        )
        .with_row_index()
        .with_columns(
            **{
                # Group HOR units based on distance.
                group_name: pl.when(pl.col("dst_behind").le(dst))
                # We assign 0 if within merge dst.
                .then(pl.lit(0))
                # Otherwise, give unique index.
                .otherwise(pl.col("index") + 1)
                # Then create run-length ID to group on.
                # Contiguous rows within distance will be grouped together.
                .rle_id()
            },
        )
        .with_columns(
            # Adjust groups in scenarios where should join group ahead or behind but given unique group.
            # B:64617 A:52416 G:1
            # B:52416 A:1357  G:2 <- This should be group 3.
            # B:1357  A:1358  G:3
            pl.when(pl.col("dst_behind").le(dst) & pl.col("dst_ahead").le(dst))
            .then(pl.col(group_name))
            .when(pl.col("dst_behind").le(dst))
            .then(pl.col(group_name).shift(1))
            .when(pl.col("dst_ahead").le(dst))
            .then(pl.col(group_name).shift(-1))
            .otherwise(pl.col(group_name))
        )
    )


def hor_array_length(
    df_stv: pl.DataFrame,
    df_rm: pl.DataFrame | None = None,
    bp_merge_units: int = DEF_BP_MERGE_UNITS,
    bp_merge_blks: int = DEF_BP_MERGE_BLKS,
    min_blk_hor_units: int = DEF_MIN_BLK_HOR_UNITS,
    min_arr_hor_units: int = DEF_MIN_ARR_HOR_UNITS,
    min_arr_len: int = DEF_MIN_ARR_LEN,
    min_arr_prop: float = DEF_MIN_ARR_PROP,
    *,
    output_strand: bool = True,
    allow_nonlive: bool = False,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    dfs: list[pl.DataFrame] = []
    dfs_strand: list[pl.DataFrame] = []
    for ctg_name, df_chr in df_stv.sort(by=["chrom", "chrom_st"]).group_by(
        ["chrom"], maintain_order=True
    ):
        ctg_name = ctg_name[0]
        if not allow_nonlive:
            df_chr = df_chr.filter(pl.col("name").str.contains("L"))

        df_live_hor = group_by_dst(
            df_chr,
            bp_merge_units,
            "live_group",
        ).filter(
            # Filter any live group with fewer than required number of HOR units.
            pl.col("live_group").count().over("live_group") >= min_blk_hor_units
        )
        itvs_live_hor = bed_to_itree(
            df_live_hor, lambda st, end: MergeHORData(ctg_name, 1, end - st)
        )

        # Filter to chrom
        if isinstance(df_rm, pl.DataFrame):
            itvs_chr_rm = it.IntervalTree(
                it.Interval(st, end, rtype)
                for st, end, rtype in df_rm.filter(pl.col("contig") == ctg_name)
                .select("start", "end", "type")
                .iter_rows()
            )
        else:
            itvs_chr_rm = None

        # Then check what's between our intervals before merging.
        def check_correct_merge(itv_1: it.Interval, itv_2: it.Interval) -> bool:
            # If no repeatmasker tracks, no second check.
            if not itvs_chr_rm:
                return True
            itv_between = it.Interval(itv_1.end, itv_2.begin)
            # Prevent costly itree lookup if only small interval.
            if itv_between.length() <= 1:
                return True

            rm_ovl = itvs_chr_rm.overlap(itv_between)

            if not rm_ovl:
                return True
            repeat_count: Counter[str] = Counter()
            for ovl in rm_ovl:
                repeat_count[ovl.data] += ovl.overlap_size(itv_between)
            # Should never KeyError as we return early if rm_ovl is empty.
            most_common_repeat, _ = repeat_count.most_common(1)[0]

            # Only merge if most common repeat is allowed.
            return most_common_repeat not in DEF_MERGE_RBLACKLIST

        merged_itvs = merge_itvs(
            itvs_live_hor.iter(),
            dst=bp_merge_blks,
            fn_cmp=check_correct_merge,
            fn_merge_itv=merge_hor_unit_itvs,
        )
        if output_strand:
            # Group HOR units by strand into blocks
            # Requiring at least the min_blk_hor_units per strand block.
            df_live_hor = (
                group_by_dst(
                    df_live_hor.with_columns(
                        strand_group=pl.col("strand").rle_id()
                    ).filter(
                        pl.col("strand_group").count().over("strand_group")
                        >= min_blk_hor_units
                    ),
                    # Allow grouping by merge_blks as this is final group check before splitting by strand_group
                    bp_merge_blks,
                    "live_group",
                )
                # Take both strand and distance into consideration.
                .with_columns(
                    strand_group=pl.col("strand").rle_id() + pl.col("live_group")
                )
            )
            for _, df_strand_group in df_live_hor.group_by(["strand_group"]):
                strand = df_strand_group.get_column("strand")[0]
                itvs_strand = bed_to_itree(
                    df_strand_group, lambda st, end: MergeHORData(strand, 1, end - st)
                )
                merged_strand_itvs = merge_itvs(
                    itvs_strand.iter(),
                    dst=bp_merge_blks,
                    fn_cmp=check_correct_merge,
                    fn_merge_itv=merge_hor_unit_itvs,
                )
                df_strand = pl.DataFrame(
                    (
                        (
                            ctg_name,
                            itv.begin,
                            itv.end,
                            itv.length(),
                            itv.data[1],
                            itv.data[2] / itv.length(),
                            strand,
                        )
                        for itv in merged_strand_itvs
                    ),
                    orient="row",
                    schema=DEF_OUTPUT_BED_COLS_STRAND,
                ).filter(
                    (pl.col("score") >= min_arr_hor_units)
                    & (pl.col("prop") >= min_arr_prop)
                    & (pl.col("name") >= min_arr_len)
                )
                dfs_strand.append(df_strand)

        df = (
            pl.DataFrame(
                (
                    (
                        ctg_name,
                        itv.begin,
                        itv.end,
                        itv.length(),
                        itv.data[1],
                        itv.data[2] / itv.length(),
                    )
                    for itv in merged_itvs
                ),
                orient="row",
                schema=DEF_OUTPUT_BED_COLS,
            )
            # Require that array has at least n merged HOR units.
            .filter(
                (pl.col("score") >= min_arr_hor_units)
                & (pl.col("prop") >= min_arr_prop)
                & (pl.col("name") >= min_arr_len)
            )
        )

        dfs.append(df)

    df_all = pl.concat(dfs).sort(by=["chrom", "chrom_st"])
    df_all_strand = (
        pl.concat(dfs_strand)
        if output_strand
        else pl.DataFrame(schema=DEF_OUTPUT_BED_COLS_STRAND)
    ).sort(by=["chrom", "chrom_st"])
    return df_all, df_all_strand



import re
import sys
import argparse
import polars as pl
import editdistance
from loguru import logger
from typing import TextIO, TYPE_CHECKING, Any

from .repeat_jaccard_index import jaccard_index, get_contig_similarity_by_jaccard_index
from .repeat_edit_dst import get_contig_similarity_by_edit_dst
from .constants import (
    CHROMOSOMES_13_21,
    CHROMOSOMES_14_22,
    MAX_ALR_LEN_THR,
    RGX_CHR,
    EDGE_LEN,
    EDGE_PERC_ALR_THR,
    DST_PERC_THR,
)
from .orientation import Orientation
from .reference import split_ref_rm_input_by_contig
from .reader import read_repeatmasker_output
from .partial_cen import is_partial_centromere


if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def join_summarize_results(
    df_partial_contig_res: pl.DataFrame,
    df_jaccard_index_res: pl.DataFrame,
    df_edit_distance_res: pl.DataFrame,
    df_edit_distance_same_chr_res: pl.DataFrame,
    *,
    reference_prefix: str,
) -> pl.DataFrame:
    df_joined = (
        df_partial_contig_res.join(
            df_jaccard_index_res.join(df_edit_distance_res, on="contig")
            .group_by("contig")
            .first(),
            on="contig",
            how="left",
        )
        # Add default ort per contig.
        .join(df_edit_distance_same_chr_res, on="contig", how="left")
        .with_columns(ctg_chr=pl.col("contig").str.extract(RGX_CHR.pattern))
    )

    return (
        df_joined.select(
            contig=pl.col("contig"),
            # Extract chromosome name.
            # Both results must concur.
            final_contig=pl.when(pl.col("ref") == pl.col("ref_right"))
            .then(pl.col("ref").str.extract(RGX_CHR.pattern))
            .otherwise(pl.col("contig").str.extract(RGX_CHR.pattern)),
            # Only use orientation if both agree. Otherwise, replace with best same chr ort.
            reorient=pl.when(pl.col("ref") == pl.col("ref_right"))
            .then(pl.col("ort"))
            .otherwise(None)
            .fill_null(pl.col("ort_same_chr")),
            partial=pl.col("partial"),
        )
        # Replace chr name in original contig.
        .with_columns(
            final_contig=pl.col("contig").str.replace(
                RGX_CHR.pattern, pl.col("final_contig")
            ),
            # Never reorient if reference.
            reorient=pl.when(pl.col("contig").str.starts_with(reference_prefix))
            .then(
                pl.col("reorient").str.replace(
                    str(Orientation.Reverse), str(Orientation.Forward)
                )
            )
            .otherwise(pl.col("reorient")),
        )
        # Take only first row per contig.
        .group_by("contig", maintain_order=True)
        .first()
    )


def check_cens_status(
    input_rm: TextIO,
    output: TextIO,
    reference_rm: str,
    *,
    reference_prefix: str,
    dst_perc_thr: float = DST_PERC_THR,
    edge_perc_alr_thr: float = EDGE_PERC_ALR_THR,
    edge_len: int = EDGE_LEN,
    max_alr_len_thr: int = MAX_ALR_LEN_THR,
    restrict_13_21: bool = False,
    restrict_14_22: bool = False,
) -> int:
    df_ctg = read_repeatmasker_output(input_rm)
    df_ref = read_repeatmasker_output(reference_rm).filter(
        pl.col("contig").str.starts_with(reference_prefix)
    )

    contigs, refs, dsts, orts = [], [], [], []
    jcontigs, jrefs, jindex = [], [], []
    pcontigs, pstatus = [], []

    # Split ref dataframe by chromosome.
    df_ref_grps = dict(split_ref_rm_input_by_contig(df_ref))
    logger.info(f"Read {len(df_ref_grps)} reference dataframes.")

    for ctg, df_ctg_grp in df_ctg.group_by(["contig"]):
        ctg_name = ctg[0]
        logger.info(f"Evaluating {ctg_name} with {df_ctg_grp.shape[0]} repeats.")

        mtch_chr_name = re.search(RGX_CHR, ctg_name)
        if not mtch_chr_name:
            continue
        chr_name = mtch_chr_name.group()

        # Check if partial ctg.
        pcontigs.append(ctg_name)
        is_partial = is_partial_centromere(
            df_ctg_grp,
            edge_len=edge_len,
            edge_perc_alr_thr=edge_perc_alr_thr,
            max_alr_len_thr=max_alr_len_thr,
        )
        pstatus.append(is_partial)

        for ref_name, ref_ctg in df_ref_grps.items():
            df_ref_grp = ref_ctg.df

            # Special case for 13 and 21 and 14 and 22.
            if (
                (chr_name in CHROMOSOMES_13_21 and ref_ctg.chr not in CHROMOSOMES_13_21)
                and restrict_13_21
            ) or (
                (chr_name in CHROMOSOMES_14_22 and ref_ctg.chr not in CHROMOSOMES_14_22)
                and restrict_14_22
            ):
                continue

            dst_fwd = editdistance.eval(
                df_ref_grp["type"].to_list(),
                df_ctg_grp["type"].to_list(),
            )
            dst_rev = editdistance.eval(
                df_ref_grp["type"].to_list(),
                df_ctg_grp["type"].reverse().to_list(),
            )

            repeat_type_jindex = jaccard_index(
                set(df_ref_grp["type"]), set(df_ctg_grp["type"])
            )
            jcontigs.append(ctg_name)
            jrefs.append(ref_name)
            jindex.append(repeat_type_jindex)

            contigs.append(ctg_name)
            contigs.append(ctg_name)
            refs.append(ref_name)
            refs.append(ref_name)
            orts.append(str(Orientation.Forward))
            orts.append(str(Orientation.Reverse))
            dsts.append(dst_fwd)
            dsts.append(dst_rev)

    df_jaccard_index_res = get_contig_similarity_by_jaccard_index(
        jcontigs, jrefs, jindex
    )
    (
        df_filter_edit_distance_res,
        df_filter_ort_same_chr_res,
    ) = get_contig_similarity_by_edit_dst(
        contigs, refs, dsts, orts, dst_perc_thr=dst_perc_thr
    )
    df_partial_contig_res = pl.DataFrame({"contig": pcontigs, "partial": pstatus})

    res = join_summarize_results(
        df_partial_contig_res=df_partial_contig_res,
        df_jaccard_index_res=df_jaccard_index_res,
        df_edit_distance_res=df_filter_edit_distance_res,
        df_edit_distance_same_chr_res=df_filter_ort_same_chr_res,
        reference_prefix=reference_prefix,
    )

    res.write_csv(output, include_header=False, separator="\t")
    logger.info("Finished checking centromeres.")

    return 0


def add_status_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "status",
        description="Determines if centromeres are incorrectly oriented/mapped with respect to a reference.",
    )
    ap.add_argument(
        "-i",
        "--input",
        help="Input RepeatMasker output. Should contain contig reference. Expects no header.",
        type=argparse.FileType("rb"),
        required=True,
    )
    ap.add_argument(
        "-o",
        "--output",
        help="List of contigs with actions required to fix.",
        default=sys.stdout,
        type=argparse.FileType("wt"),
    )
    ap.add_argument(
        "-r",
        "--reference",
        required=True,
        type=str,
        help="Reference RM dataframe.",
    )
    ap.add_argument(
        "--dst_perc_thr",
        default=DST_PERC_THR,
        type=float,
        help="Edit distance percentile threshold. Lower is more stringent.",
    )
    ap.add_argument(
        "--edge_perc_alr_thr",
        default=EDGE_PERC_ALR_THR,
        type=float,
        help="Percent ALR on edges of contig to be considered a partial centromere.",
    )
    ap.add_argument(
        "--edge_len",
        default=EDGE_LEN,
        type=int,
        help="Edge len to calculate edge_perc_alr_thr.",
    )
    ap.add_argument(
        "--max_alr_len_thr",
        default=MAX_ALR_LEN_THR,
        type=int,
        help="Length of largest ALR needed in a contig to not be considered a partial centromere.",
    )
    ap.add_argument(
        "--reference_prefix", default="chm13", type=str, help="Reference prefix."
    )
    ap.add_argument(
        "--restrict_13_21",
        action="store_true",
        help="Restrict mapping to chromosomes 13 and 21 for chr13 and chr21 contigs.",
    )
    ap.add_argument(
        "--restrict_14_22",
        action="store_true",
        help="Restrict mapping to chromosomes 14 and 22 for chr14 and chr22 contigs.",
    )

    return None



import sys
import argparse

from typing import TYPE_CHECKING, Any, TextIO

from .estimate_length import hor_array_length
from .constants import (
    DEF_MIN_BLK_HOR_UNITS,
    DEF_MIN_ARR_HOR_UNITS,
    DEF_MIN_ARR_LEN,
    DEF_MIN_ARR_PROP,
    DEF_BP_MERGE_UNITS,
    DEF_BP_MERGE_BLKS,
    DEF_INPUT_BED_COLS,
    DEF_INPUT_RM_COLS,
    DEF_INPUT_RM_COL_IDX,
    DEF_OUTPUT_BED_COLS,
    DEF_OUTPUT_BED_COLS_STRAND,
)
from .io import format_and_output_lengths, read_stv, read_rm

if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_hor_length_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "length",
        description="Estimate HOR array length from stv bed file / HumAS-HMMER output.",
    )
    ap.add_argument(
        "-i",
        "--input_stv",
        help=f"Input stv row bed file produced by HumAS-HMMER and stv. Expects columns: {DEF_INPUT_BED_COLS}",
        type=argparse.FileType("rb"),
    )
    ap.add_argument(
        "-r",
        "--input_rm",
        help=f"Input tab-delimited RepeatMasker file with no header. Prevents joining across. Expects columns: {DEF_INPUT_RM_COLS} at indices {DEF_INPUT_RM_COL_IDX}.",
        type=argparse.FileType("rb"),
        default=None,
    )
    ap.add_argument(
        "-o",
        "--output",
        help=f"Output bed file with columns: {DEF_OUTPUT_BED_COLS}",
        default=sys.stdout,
        type=argparse.FileType("wt"),
    )
    ap.add_argument(
        "-s",
        "--output_strand",
        help=f"Output bed file with columns: {DEF_OUTPUT_BED_COLS}",
        default=None,
        type=str,
    )
    ap.add_argument(
        "-mu",
        "--bp_merge_units",
        help="Merge HOR units into HOR blocks within this number of base pairs.",
        type=int,
        default=DEF_BP_MERGE_UNITS,
    )
    ap.add_argument(
        "-mb",
        "--bp_merge_blks",
        help="Merge HOR blocks into HOR arrays within this number of bases pairs.",
        type=int,
        default=DEF_BP_MERGE_BLKS,
    )
    ap.add_argument(
        "-ub",
        "--min_blk_hor_units",
        help="HOR blocks must have at least n HOR units unbroken.",
        type=int,
        default=DEF_MIN_BLK_HOR_UNITS,
    )
    ap.add_argument(
        "-ua",
        "--min_arr_hor_units",
        help="Require that an HOR array have at least n HOR units.",
        type=int,
        default=DEF_MIN_ARR_HOR_UNITS,
    )
    ap.add_argument(
        "-fp",
        "--min_arr_prop",
        help="Require that an HOR array has at least this proportion of HOR by length.",
        type=float,
        default=DEF_MIN_ARR_PROP,
    )
    ap.add_argument(
        "-fl",
        "--min_arr_len",
        help="Require that an HOR array is this size in bp.",
        type=int,
        default=DEF_MIN_ARR_LEN,
    )
    ap.add_argument(
        "--allow_nonlive",
        action="store_true",
        help="Don't filter for L in name column.",
    )
    return None


def calculate_hor_length(
    infile: TextIO,
    bp_merge_units: int,
    bp_merge_blks: int,
    min_blk_hor_units: int,
    min_arr_hor_units: int,
    min_arr_len: int,
    min_arr_prop: int,
    output: TextIO,
    output_strand: str | None = None,
    rmfile: TextIO | None = None,
    allow_nonlive: bool = False,
) -> int:
    """
    Calculate HOR array length from HumAS-HMMER structural variation row output.

    ### Parameters
    `infile`
        Input bed file made from HumAS-HMMER output.
        Expects the following columns: `{chrom, chrom_st, chrom_end, hor, 0, strand, ...}`.
    `rmfile`
        Input RepeatMasker file.
        Used to prevent merging across other repeat types.
    `bp_merge_units`
        Merge HOR units into HOR blocks within this number of base pairs.
    `bp_merge_blks`
        Merge HOR blocks into HOR arrays within this number of bases pairs.
    `min_blk_hor_units`
        Grouped stv rows must have at least `n` HOR units unbroken.
    `min_arr_hor_units`
        Require that an HOR array have at least `n` HOR units.
    `min_arr_len`
        Require that an HOR array is this size in bp.
    `min_arr_prop`
        Require that an HOR array has at least this proportion of HORs by length.
    `output`
        Output bed file with HOR array lengths.
        Columns: `{chrom, chrom_st, chrom_end, length}`.
    `output_strand`
        Output bed file with HOR array lengths by strand.
        Columns: `{chrom, chrom_st, chrom_end, length, strand}`.
    `allow_nonlive`
        Don't filter for `L` character.

    ### Returns
    0 if successful.
    """
    df_stv = read_stv(infile)
    df_rm = read_rm(rmfile) if rmfile else None

    df_all_len, df_all_strand_len = hor_array_length(
        df_stv=df_stv,
        df_rm=df_rm,
        bp_merge_units=bp_merge_units,
        bp_merge_blks=bp_merge_blks,
        min_blk_hor_units=min_blk_hor_units,
        min_arr_hor_units=min_arr_hor_units,
        min_arr_len=min_arr_len,
        min_arr_prop=min_arr_prop,
        output_strand=isinstance(output_strand, str),
        allow_nonlive=allow_nonlive,
    )

    if output_strand:
        format_and_output_lengths(
            df_all_strand_len, output_strand, DEF_OUTPUT_BED_COLS_STRAND
        )

    format_and_output_lengths(df_all_len, output, DEF_OUTPUT_BED_COLS)
    return 0



import os
import pyfaidx
import argparse

from loguru import logger
from typing import TYPE_CHECKING, Any, Generator
from enum import StrEnum
from statistics import mean
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

from .estimate_identity import convertMatrixToBed, createSelfMatrix
from .read_fasta import generateKmersFromFasta


if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


class Dim(StrEnum):
    ONE = "1D"
    TWO = "2D"


def convert_2D_to_1D_ident(
    bed: list[tuple[str, int, int, str, int, int, float]],
    window: int,
    n_bins: int,
    ignore_bands: int,
) -> Generator[tuple[int, int, float], None, None]:
    # Use dictionary to avoid sparse mtx.
    aln_mtx: defaultdict[int, dict[int, float]] = defaultdict(dict)
    for line in bed:
        q_ctg, q_st, q_end, r_ctg, r_st, r_end, perID = line
        # Convert position to indices.
        x = q_st // window
        y = r_st // window
        aln_mtx[x][y] = perID

    st_idxs = list(aln_mtx.keys())
    for st_idx in st_idxs:
        st = st_idx * window + 1
        end = st + window - 1
        band_end_idx = st_idx + n_bins
        # Within the alignment matrix with a n_bins of 5 and ignore_bands of 2:
        # - '*' is the calculated aln band
        # - '+' is self aln.
        # 4 * * *   +
        # 3 * *   +
        # 2 *   +
        # 1   +
        # 0 +
        #   0 1 2 3 4
        mean_ident = mean(
            aln_mtx[x].get(y, 0.0)
            for x in range(st_idx, band_end_idx)
            for y in range(x + ignore_bands, band_end_idx)
        )
        yield st, end, mean_ident


def get_single_self_seq_ident(
    seq_id: str,
    seq: str,
    outdir: str,
    window: int,
    delta: float,
    kmer_size: int,
    ident_thr: float,
    modimizer: int,
    n_bins: int,
    ignore_bands: int,
    dim: Dim,
    round_ndigits: int | None,
) -> None:
    logger.info(f"Generating self sequence identity for {seq_id}.")
    kmers = [kmer_hash for kmer_hash in generateKmersFromFasta(seq, kmer_size)]
    mtx = createSelfMatrix(kmers, window, delta, kmer_size, ident_thr, False, modimizer)
    bed = convertMatrixToBed(mtx, window, ident_thr, seq_id, seq_id, True)
    outfile = os.path.join(outdir, f"{seq_id}.bed")

    if dim == Dim.TWO:
        logger.info(
            f"Writing 2D self sequence identity array for {seq_id} to {outfile}"
        )
        with open(outfile, "wt") as fh:
            for qname, qst, qend, rname, rst, rend, ident in bed:
                ident = round(ident, round_ndigits) if round_ndigits else ident
                fh.write(f"{qname}\t{qst}\t{qend}\t{rname}\t{rst}\t{rend}\t{ident}\n")
    else:
        logger.info(f"Converting 2D self sequence identity matrix to 1D for {seq_id}.")
        logger.info(
            f"Writing 1D self sequence identity array for {seq_id} to {outfile}"
        )
        with open(outfile, "wt") as fh:
            for st, end, ident in convert_2D_to_1D_ident(
                bed, window, n_bins, ignore_bands
            ):
                ident = round(ident, round_ndigits) if round_ndigits else ident
                fh.write(f"{seq_id}\t{st}\t{end}\t{ident}\n")

    return None


def add_self_ident_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "self-ident",
        description="Approximate 1D or 2D average self nucleotide identity via a k-mer-based containment index. Uses ModDotPlot's library.",
    )
    ap.add_argument(
        "-i",
        "--infile",
        required=True,
        type=str,
        help="Input fasta.",
    )
    ap.add_argument(
        "-o",
        "--outdir",
        type=str,
        required=True,
        help="Output directory for self-identity alignment bedfile by contig.",
    )
    ap.add_argument(
        "-x",
        "--dim",
        choices=[Dim.ONE, Dim.TWO],
        help="Dimensionality of self-identity returned.",
        default=Dim.ONE,
        type=Dim,
    )
    ap.add_argument(
        "-p", "--processes", default=4, type=int, help="Number of processes."
    )
    ap.add_argument(
        "-t", "--ident_thr", default=0.86, type=float, help="Identity threshold."
    )
    ap.add_argument("-w", "--window", default=5000, type=int, help="Window size.")
    ap.add_argument("-k", "--kmer_size", default=21, type=int, help="K-mer size.")
    ap.add_argument(
        "-d",
        "--delta",
        default=0.5,
        type=float,
        help="Fraction of neighboring partition to include in identity estimation. Must be between 0 and 1, use > 0.5 is not recommended.",
    )
    ap.add_argument(
        "-m",
        "--modimizer",
        default=1000,
        type=int,
        help="Modimizer sketch size. A lower value will reduce the number of modimizers, but will increase performance. Must be less than --window.",
    )
    # 1D params
    ap.add_argument(
        "-b",
        "--n_bins",
        default=5,
        type=int,
        help="Number of bins to calculate average sequence identity over. Only applicable if mode is 1D.",
    )
    ap.add_argument(
        "--ignore_bands",
        default=2,
        type=int,
        help="Number of bands ignored along self-identity diagonal. Only applicable if mode is 1D.",
    )
    ap.add_argument(
        "--round_ndigits",
        default=None,
        type=int,
        help="Round identity to specified ndigits.",
    )
    return None


def get_self_seq_ident(
    infile: str,
    outdir: str,
    window: int,
    delta: float,
    kmer_size: int,
    ident_thr: float,
    modimizer: int,
    n_bins: int,
    ignore_bands: int,
    processes: int,
    dim: Dim,
    round_ndigits: int | None,
):
    os.makedirs(outdir, exist_ok=True)

    seq = pyfaidx.Fasta(infile)
    with ProcessPoolExecutor(max_workers=processes) as pool:
        _ = pool.map(
            get_single_self_seq_ident,
            *zip(
                *[
                    (
                        sec_rec.name,
                        str(sec_rec),
                        outdir,
                        window,
                        delta,
                        kmer_size,
                        ident_thr,
                        modimizer,
                        n_bins,
                        ignore_bands,
                        dim,
                        round_ndigits,
                    )
                    for sec_rec in seq
                ]
            ),
        )



import argparse
from typing import Any, TYPE_CHECKING

from .status.cli import add_status_cli, check_cens_status
from .length.cli import add_hor_length_cli, calculate_hor_length
from .nonredundant.cli import add_nonredundant_cli, get_nonredundant_cens
from .entropy.cli import add_entropy_cli, calculate_windowed_shannon_index
from .self_ident.cli import add_self_ident_cli, get_self_seq_ident


if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def main() -> int:
    ap = argparse.ArgumentParser(description="Centromere statistics toolkit.")
    sub_ap = ap.add_subparsers(dest="cmd")
    add_status_cli(sub_ap)
    add_hor_length_cli(sub_ap)
    add_nonredundant_cli(sub_ap)
    add_entropy_cli(sub_ap)
    add_self_ident_cli(sub_ap)

    args = ap.parse_args()

    if args.cmd == "status":
        return check_cens_status(
            args.input,
            args.output,
            args.reference,
            reference_prefix=args.reference_prefix,
            dst_perc_thr=args.dst_perc_thr,
            edge_len=args.edge_len,
            edge_perc_alr_thr=args.edge_perc_alr_thr,
            max_alr_len_thr=args.max_alr_len_thr,
            restrict_13_21=args.restrict_13_21,
            restrict_14_22=args.restrict_14_22,
        )
    elif args.cmd == "length":
        return calculate_hor_length(
            infile=args.input_stv,
            rmfile=args.input_rm,
            bp_merge_units=args.bp_merge_units,
            bp_merge_blks=args.bp_merge_blks,
            min_blk_hor_units=args.min_blk_hor_units,
            min_arr_hor_units=args.min_arr_hor_units,
            min_arr_len=args.min_arr_len,
            min_arr_prop=args.min_arr_prop,
            output=args.output,
            output_strand=args.output_strand,
            allow_nonlive=args.allow_nonlive,
        )
    elif args.cmd == "nonredundant":
        return get_nonredundant_cens(
            args.infile_left,
            args.infile_right,
            args.outfile_left,
            args.outfile_right,
            args.outfile_both,
            args.duplicates_left,
            args.duplicates_right,
            bp_diff=args.diff_bp,
        )
    elif args.cmd == "entropy":
        return calculate_windowed_shannon_index(
            args.input,
            args.outdir,
            args.window,
            args.ignore_repeats,
            # args.cores,
            omit_plot=args.omit_plot,
        )
    elif args.cmd == "self-ident":
        return get_self_seq_ident(
            args.infile,
            args.outdir,
            args.window,
            args.delta,
            args.kmer_size,
            args.ident_thr,
            args.modimizer,
            args.n_bins,
            args.ignore_bands,
            args.processes,
            args.dim,
            args.round_ndigits,
        )
    else:
        raise ValueError(f"Unknown command: {args.cmd}")


if __name__ == "__main__":
    raise SystemExit(main())



import os
import math
import argparse
import polars as pl
import matplotlib.pyplot as plt
from typing import Generator, TextIO, TYPE_CHECKING, Any
from collections import Counter
# from concurrent.futures import ProcessPoolExecutor

from matplotlib.colors import LinearSegmentedColormap
from intervaltree import Interval, IntervalTree
from loguru import logger

from .constants import DEF_FILTER_RP, DEF_BED9_COLS, DEF_WINDOW_SIZE
from ..common import merge_itvs

if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def get_shannon_index_itvs(
    df: pl.DataFrame,
    window_size: int = DEF_WINDOW_SIZE,
    filter_repeats: set[str] | None = None,
) -> Generator[Interval, None, None]:
    """
    Calculate windowed shannon index from repeat content.

    # Args
    * df
            * Dataframe with columns: `["chromStart", "chromEnd", and "name"]`
    * window_size
            * Window size in bases to calculate shannon index over.
            * By default, 5000 bp.
    * filter_repeats
            * Repeats to filter out.
            * By default, `SAR` and `Simple_repeat`

    # Returns
    Generator of `Interval`s with shannon index in `data` attribute.
    """
    if not filter_repeats:
        filter_repeats = DEF_FILTER_RP

    complete_interval = Interval(df["chromStart"].min(), df["chromEnd"].max())
    windows = list(range(complete_interval.begin, complete_interval.end, window_size))
    intervals_windows = IntervalTree()
    for i, start in enumerate(windows):
        try:
            stop = windows[i + 1]
        except IndexError:
            stop = df["chromEnd"].max()
        intervals_windows.addi(start, stop)

    intervals_df = IntervalTree.from_tuples(
        (row["chromStart"], row["chromEnd"], row["name"])
        for row in df.filter(~pl.col("name").is_in(filter_repeats)).iter_rows(
            named=True
        )
    )
    intervals_clipped_windows = []

    for wd in sorted(intervals_windows.iter()):
        overlap = intervals_df.overlap(wd)
        intervals_overlap: IntervalTree = IntervalTree(overlap)
        # Subtract left and right.
        intervals_overlap.chop(complete_interval.begin, wd.begin)
        intervals_overlap.chop(wd.end, complete_interval.end)

        intervals_clipped_windows.append(intervals_overlap)

    for intervals_window in intervals_clipped_windows:
        rp_cnts: Counter[str] = Counter()
        if intervals_windows.is_empty():
            continue
        window_st = intervals_window.begin()
        window_end = intervals_window.end()

        if window_st == 0:
            continue

        for interval_rp in intervals_window:
            rp_cnts[interval_rp.data] += interval_rp.length()
        rp_prop = [cnt / rp_cnts.total() for _, cnt in rp_cnts.items()]
        # https://www.statology.org/shannon-diversity-index/
        num_rp = len(rp_cnts)
        if num_rp <= 1:
            sh_idx = 0.0
        else:
            sh_entropy = -sum(p * math.log(p) for p in rp_prop)
            sh_idx = float(sh_entropy / math.log(num_rp))

        yield Interval(window_st, window_end, round(sh_idx, 3))


def calculate_single_windowed_shannon_index(
    df_group: tuple[tuple[object, ...], pl.DataFrame],
    outdir: str | None,
    window: int,
    ignore_repeats: list[str],
    *,
    omit_plot: bool,
) -> pl.DataFrame | None:
    """
    Calculate windowed shannon index for a chrom `DataFrame` group.

    # Args
    * df_group
            * DataFrame group from `DataFrame.partition()` or `DataFrame.group_by()`
    * outdir
            * Output directory. If `None`, return `DataFrame`
    * window
            * Window size
    * ignore_regions
            * Regions to ignore
    * omit_plot
            * Do not generate plots.

    # Returns
    `DataFrame` if no `outdir`.
    """
    grp, df = df_group
    chrom = grp[0]
    logger.info(f"Calculating Shannon index for {chrom}")
    itvs = list(get_shannon_index_itvs(df, window, filter_repeats=set(ignore_repeats)))
    merged_itvs = merge_itvs(
        itvs,
        dst=1,
        # Merge if index equal.
        fn_cmp=lambda x, y: x.data == y.data,
        fn_merge_itv=lambda x, y: Interval(x.begin, y.end, x.data),
    )
    # Scale colors based on index
    cmap = LinearSegmentedColormap.from_list("", ["red", "orange", "green"])
    df_entropy = pl.DataFrame(
        [
            (
                chrom,
                i.begin,
                i.end,
                "shannon_index",
                i.data,
                "+",
                i.begin,
                i.end,
                # Convert scaled color to rgb
                ",".join(str(round(clr * 255)) for clr in cmap(i.data)[0:-1]),
            )
            for i in merged_itvs
        ],
        schema=DEF_BED9_COLS,
        orient="row",
    )
    if not outdir:
        return df_entropy

    if not omit_plot:
        logger.info(f"Generating plot for {chrom}")
        plt.clf()
        _, ax = plt.subplots(figsize=(20, 10))

        # Draw repeat colors
        for r in df.iter_rows(named=True):
            if r["name"] == "ALR/Alpha" and not r.get("rgb"):
                color = "red"
            else:
                color = r.get("rgb")
            ax.axvspan(
                xmin=r["chromStart"], xmax=r["chromEnd"], facecolor=color, alpha=0.3
            )

        # Plot original values.
        begin, entropy = zip(*[(itv.begin, itv.data) for itv in itvs])
        ax.plot(begin, entropy, color="black")
        ax.fill_between(begin, entropy, color="black")
        ax.margins(x=0, y=0)

        plt.title(f"{chrom} ({window=:,}bp)")
        plt.xlabel("Position")
        plt.ylabel("Shannon index")
        plt.minorticks_on()
        plt.savefig(os.path.join(outdir, f"{chrom}.png"), bbox_inches="tight")
        plt.close()

    df_entropy.write_csv(
        os.path.join(outdir, f"{chrom}.bed"), separator="\t", include_header=False
    )
    return None


def calculate_windowed_shannon_index(
    infile: TextIO,
    outdir: str,
    window_size: int,
    ignore_repeats: list[str],
    # cores: int,
    *,
    omit_plot: bool,
) -> int:
    df_all = pl.read_csv(infile, separator="\t", has_header=False)
    n_cols = df_all.shape[1]
    df_all = df_all.rename(
        dict(zip([f"column_{c}" for c in range(1, n_cols)], DEF_BED9_COLS[:n_cols]))
    )
    os.makedirs(outdir, exist_ok=True)

    df_grps = df_all.partition_by(["chrom"], as_dict=True, maintain_order=True).items()
    for grp in df_grps:
        calculate_single_windowed_shannon_index(
            grp, outdir, window_size, ignore_repeats, omit_plot=omit_plot
        )

    # with ProcessPoolExecutor(cores) as pool:
    # _ = pool.map(
    #     calculate_single_windowed_shannon_index,
    #     *zip(
    #         *[
    #             (grp, outdir, window_size, ignore_regions)
    #             for grp in df_all.partition_by(["chrom"], as_dict=True, maintain_order=True).items()
    #         ]
    #     )
    # )
    return 0


def add_entropy_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "entropy",
        description="Calculate shannon index across a region from RepeatMasker repeats.",
    )
    ap.add_argument(
        "-i",
        "--input",
        required=True,
        help=f"Repeatmasker bedfile. Expects {DEF_BED9_COLS}. The 'name' column should correspond to the repeat name.",
    )
    ap.add_argument(
        "-w",
        "--window",
        type=int,
        default=DEF_WINDOW_SIZE,
        help=f"Window size. Default: {DEF_WINDOW_SIZE}",
    )
    ap.add_argument(
        "-o",
        "--outdir",
        type=str,
        required=True,
        help=(
            "Output dir. Will produce a BED9 file where 'score' corresponds to the Shannon index. "
            "The plot visualize this index across the given repeat region."
        ),
    )
    # ap.add_argument("-c", "--cores", type=int, default=4, help="Number of cores to use.")
    ap.add_argument(
        "--ignore_repeats",
        nargs="*",
        default=[],
        help=f"Repeat types to ignore in calculation. Default: {DEF_FILTER_RP}",
    )
    ap.add_argument("--omit_plot", action="store_true", help="Omit plot.")

    return None



# ModDotPlot
# https://github.com/marbl/ModDotPlot/commit/0f593a7b7b317cdfc00ef350491e17239eda594f
from typing import Generator
import pyfaidx
import mmh3

tab_b = bytes.maketrans(b"ACTG", b"TGAC")


def generateKmersFromFasta(seq: str, k: int) -> Generator[int, None, None]:
    n = len(seq)
    for i in range(n - k + 1):
        # Remove case sensitivity
        kmer = seq[i : i + k].upper()
        fh = mmh3.hash(kmer)

        # Calculate reverse complement hash directly without the need for translation
        rc = mmh3.hash(kmer[::-1].translate(tab_b))

        yield fh if fh < rc else rc


def readKmersFromFile(
    filename: str, ksize: int
) -> Generator[tuple[str, list[int]], None, None]:
    """
    Given a filename and an integer k, returns a list of all k-mers found in the sequences in the file.
    """
    seq = pyfaidx.Fasta(filename)

    for seq_rec in seq:
        kmers_for_seq = [
            kmer_hash for kmer_hash in generateKmersFromFasta(str(seq_rec), ksize)
        ]
        yield seq_rec.name, kmers_for_seq



DEF_WINDOW_SIZE = 100_000
DEF_BED9_COLS = (
    "chrom",
    "chromStart",
    "chromEnd",
    "name",
    "score",
    "strand",
    "thickStart",
    "thickEnd",
    "itemRgb",
)
DEF_FILTER_RP = {"SAR", "Simple_repeat"}



from enum import StrEnum, auto
import polars as pl


class Side(StrEnum):
    Left = auto()
    Right = auto()


BP_DIFF = 1000
IO_COLS = ["ctg", "start", "end", "length"]


def select_exprs(side: Side = Side.Left) -> dict[str, pl.Expr]:
    suffix = "_right" if side == Side.Right else ""
    return {
        f"ctg{suffix}": pl.col(f"sample{suffix}")
        + "_"
        + pl.when(pl.col(f"rc{suffix}"))
        .then("rc-" + pl.col(f"chr{suffix}"))
        .otherwise(pl.col(f"chr{suffix}"))
        + "_"
        + pl.col(f"ctg{suffix}"),
        f"start{suffix}": pl.col(f"start{suffix}"),
        f"end{suffix}": pl.col(f"end{suffix}"),
        f"length{suffix}": pl.col(f"length{suffix}"),
    }



# ModDotPlot
# https://github.com/marbl/ModDotPlot/commit/50ecda4eff91acd00584090afd380d4a355be7aa
import math
import numpy as np
import mmh3


def removeAmbiguousBases(mod_list: set[int], k: int) -> set[int]:
    # Ambiguous IUPAC codes
    bases_to_remove = ["R", "Y", "M", "K", "S", "W", "H", "B", "V", "D", "N"]
    kmers_to_remove = set()
    for i in range(len(bases_to_remove)):
        result_string = str(bases_to_remove[i]) * k
        kmers_to_remove.add(mmh3.hash(result_string))
    mod_set = set(mod_list)
    # Remove homopolymers of ambiguous nucleotides
    mod_set.difference_update(kmers_to_remove)
    return mod_set


def createSelfMatrix(
    sequence: list[int],
    window_size: int,
    delta: float,
    k: int,
    identity: float,
    ambiguous: bool,
    modimizer: int,
) -> np.ndarray:
    """
    Create self-identity matrix.

    Args:
    * sequence
            * Sequence as list of mmh3 kmers.
    * window_size
            * Window size.
    * delta
            * Fraction of neighboring partition to include in identity estimation. Must be between 0 and 1, use > 0.5 is not recommended.
    * k
            * kmer length
    * identity
            * Identity cutoff threshold.
    * ambiguous
            * Preserve diagonal when handling strings of ambiguous homopolymers (eg. long runs of N's).
    * modimizer
            * Modimizer sketch size.
            * A lower value will reduce the number of modimizers, but will increase performance.
            * Must be less than window size.

    Returns
    * 2D ndarray of identity values.
    """
    sequence_length = len(sequence)
    seq_sparsity = round(window_size / modimizer)
    if seq_sparsity <= modimizer:
        seq_sparsity = 2 ** int(math.log2(seq_sparsity))
    else:
        seq_sparsity = 2 ** (int(math.log2(seq_sparsity - 1)) + 1)
    sketch_size = round(window_size / seq_sparsity)

    no_neighbors = partitionOverlaps(sequence, window_size, 0, sequence_length, k)
    if delta > 0:
        neighbors = partitionOverlaps(sequence, window_size, delta, sequence_length, k)
    else:
        neighbors = no_neighbors

    neighbors_mods = convertToModimizers(
        neighbors, seq_sparsity, ambiguous, k, sketch_size
    )
    no_neighbors_mods = convertToModimizers(
        no_neighbors, seq_sparsity, ambiguous, k, sketch_size
    )
    matrix = selfContainmentMatrix(
        no_neighbors_mods, neighbors_mods, k, identity, ambiguous
    )
    return matrix


def partitionOverlaps(
    lst: list[int], win: int, delta: float, seq_len: int, k: int
) -> list[list[int]]:
    kmer_list = []
    kmer_to_genomic_coordinate_offset = win - k + 1
    delta_offset = win * delta

    # set the first window to contain win - k + 1 kmers.
    starting_end_index = int(round(kmer_to_genomic_coordinate_offset + delta_offset))
    kmer_list.append(lst[0:starting_end_index])
    counter = win - k + 1

    # set normal windows
    while counter <= (seq_len - win):
        start_index = counter + 1
        end_index = win + counter + 1
        delta_start_index = int(round(start_index - delta_offset))
        delta_end_index = int(round(end_index + delta_offset))
        if delta_end_index > seq_len:
            delta_end_index = seq_len
        try:
            kmer_list.append(lst[delta_start_index:delta_end_index])
        except Exception as e:
            print(e)
            kmer_list.append(lst[delta_start_index:seq_len])
        counter += win

    # set the last window to get the remainder
    if counter <= seq_len - 2:
        final_start_index = int(round(counter + 1 - delta_offset))
        kmer_list.append(lst[final_start_index:seq_len])

    # Test that last value was added on correctly

    assert kmer_list[-1][-1] == lst[-1]
    return kmer_list


def populateModimizers(
    partition: list[int], sparsity: int, ambiguous: bool, expectation: int, k: int
) -> set[int]:
    mod_set = set()
    for kmer in partition:
        if kmer % sparsity == 0:
            mod_set.add(kmer)
    if not ambiguous:
        mod_set = removeAmbiguousBases(mod_set, k)
    if (len(mod_set) < round(expectation / 2)) and (sparsity > 1):
        populateModimizers(partition, sparsity // 2, ambiguous, expectation, k)
    return mod_set


def convertToModimizers(
    kmer_list: list[list[int]], sparsity: int, ambiguous: bool, k: int, expectation: int
) -> list[set[int]]:
    mod_total = []
    for partition in kmer_list:
        mod_set = populateModimizers(partition, sparsity, ambiguous, expectation, k)
        mod_total.append(mod_set)
    return mod_total


def convertMatrixToBed(
    matrix: np.ndarray,
    window_size: int,
    id_threshold: float,
    x_name: str,
    y_name: str,
    self_identity: bool,
) -> list[tuple[str, int, int, str, int, int, float]]:
    """
    # Returns
    * "#query_name"
    * "query_start"
    * "query_end"
    * "reference_name"
    * "reference_start"
    * "reference_end"
    * "perID_by_events"
    """
    bed = []
    rows, cols = matrix.shape
    for x in range(rows):
        for y in range(cols):
            value = matrix[x, y]
            if (not self_identity) or (self_identity and x <= y):
                if value >= id_threshold / 100:
                    start_x = x * window_size + 1
                    end_x = (x + 1) * window_size
                    start_y = y * window_size + 1
                    end_y = (y + 1) * window_size

                    bed.append(
                        (
                            x_name,
                            int(start_x),
                            int(end_x),
                            y_name,
                            int(start_y),
                            int(end_y),
                            float(value),
                        )
                    )
    return bed


def binomial_distance(containment_value: float, kmer_value: int) -> float:
    """
    Calculate the binomial distance based on containment and kmer values.

    Args:
        containment_value (float): The containment value.
        kmer_value (int): The k-mer value.

    Returns:
        float: The binomial distance.
    """
    return math.pow(containment_value, 1.0 / kmer_value)


def containment_neighbors(
    set1: set[int],
    set2: set[int],
    set3: set[int],
    set4: set[int],
    identity: float,
    k: int,
) -> float:
    """
    Calculate the containment neighbors based on four sets and an identity threshold.

    Args:
    * set1:
            * The first set.
    * set2:
            * The second set.
    * set3:
            * The third set.
    * set4:
            * The fourth set.
    * identity:
            * The identity threshold.
    * k:
            * Kmer value.

    Returns:
    * The containment neighbors value.
    """
    len_a = len(set1)
    len_b = len(set2)

    intersection_a_b_prime = len(set1 & set4)
    if len_a != 0:
        containment_a_b_prime = intersection_a_b_prime / len_a
    else:
        # If len_a is zero, handle it by setting containment_a_b_prime to a default value
        containment_a_b_prime = 0

    if binomial_distance(containment_a_b_prime, k) < identity / 100:
        return 0.0

    else:
        intersection_a_prime_b = len(set2 & set3)
        if len_b != 0:
            containment_a_prime_b = intersection_a_prime_b / len_b
        else:
            # If len_a is zero, handle it by setting containment_a_b_prime to a default value
            containment_a_prime_b = 0

        return max(containment_a_b_prime, containment_a_prime_b)


def selfContainmentMatrix(
    mod_set: list[set],
    mod_set_neighbors: list[set],
    k: int,
    identity: float,
    ambiguous: bool,
) -> np.ndarray:
    """
    Create a self-containment matrix based on containment similarity calculations.

    Args:
    * mod_set
            * A list of sets representing elements.
    * mod_set_neighbors
            * A list of sets representing neighbors for each element.
    * k
            * A parameter for containment similarity calculation.

    Returns:
        np.ndarray: A NumPy array representing the self-containment matrix.
    """
    n = len(mod_set)
    containment_matrix = np.empty((n, n))

    for w in range(n):
        containment_matrix[w, w] = 100.0
        if len(mod_set[w]) == 0 and not ambiguous:
            containment_matrix[w, w] = 0

        for r in range(w + 1, n):
            c_hat = binomial_distance(
                containment_neighbors(
                    mod_set[w],
                    mod_set[r],
                    mod_set_neighbors[w],
                    mod_set_neighbors[r],
                    identity,
                    k,
                ),
                k,
            )
            containment_matrix[r, w] = c_hat * 100.0
            containment_matrix[w, r] = c_hat * 100.0

    return containment_matrix



import polars as pl
from typing import TextIO
from .constants import RM_COLS, RM_COL_IDX


def read_repeatmasker_output(input_path: TextIO | str) -> pl.DataFrame:
    return pl.read_csv(
        input_path,
        separator="\t",
        has_header=False,
        columns=RM_COL_IDX,
        new_columns=RM_COLS,
        truncate_ragged_lines=True,
    ).with_columns(dst=pl.col("end") - pl.col("start"))



import re


RM_COLS = [
    "contig",
    "start",
    "end",
    "type",
    "rClass",
]
RM_COL_IDX = [4, 5, 6, 9, 10]
CHROMOSOMES_13_21 = {
    "chr13",
    "chr21",
}
CHROMOSOMES_14_22 = {
    "chr14",
    "chr22",
}

RGX_CHR = re.compile(r"(chr[0-9XY]+)")
DST_PERC_THR = 0.3
EDGE_LEN = 100_000
EDGE_PERC_ALR_THR = 0.7
MAX_ALR_LEN_THR = 200_000
REPEAT_SPLIT_LEN = 2_000



import sys
import argparse
from typing import Any, TextIO, TYPE_CHECKING

import polars as pl
from loguru import logger

from .constants import BP_DIFF, IO_COLS, Side, select_exprs


if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
    RowVals = tuple[Any, ...]
else:
    SubArgumentParser = Any


def read_as_hor_length_tsv(file: str | TextIO) -> pl.DataFrame:
    """
    Read AS-HOR array length TSV file. Groups by ctg and gets aggregated length before formatting.
    * Expects columns: `["ctg", "start", "end", "length"]`
    * The `ctg` column must contain info for `["sample", "chr", "ctg"]` and be `'_'` delimited.
    * `ctg` should start with the haplotype information. Either 1 or 2.

    Example:
    * `HG01114_rc-chr1_h2tg000002l#1-130810013:121319346-129631944`
    * `HG01573_rc-chr1_haplotype1-0000024:121168122-126852171`
    """
    return (
        pl.read_csv(
            file,
            separator="\t",
            has_header=False,
            new_columns=IO_COLS,
        )
        .group_by("ctg")
        .agg(
            pl.col("start").min(), pl.col("end").max(), pl.sum("length").alias("length")
        )
        .with_columns(og_ctg=pl.col("ctg"), ctg=pl.col("ctg").str.split_exact("_", 2))
        .unnest("ctg")
        .rename({"field_0": "sample", "field_1": "chr", "field_2": "ctg"})
        .with_columns(
            # TODO: Handle mat or pat.
            hap=pl.col("ctg")
            .str.extract(r"(^h.*?\d)", 1)
            .replace({"haplotype1": "h1", "haplotype2": "h2"})
            .fill_null("unassigned"),
            rc=pl.col("chr").str.contains("rc-"),
            chr=pl.col("chr").str.replace("rc-", ""),
        )
    )


def get_nonredundant_cens(
    infile_left: TextIO,
    infile_right: TextIO,
    outfile_left: str,
    outfile_right: str,
    outfile_both: str,
    outfile_dupe_left: str,
    outfile_dupe_right: str,
    *,
    bp_diff: int = BP_DIFF,
):
    # Read AS-HOR length dataframe.
    # Calculate cumulative AS-HOR array length per centromere
    # Parse haplotype, chr, sample, and ctg_num_coord.
    df_left_og_fmt = read_as_hor_length_tsv(infile_left)
    df_right_og_fmt = read_as_hor_length_tsv(infile_right)

    df_merged_og_fmt = pl.concat(
        [
            df_left_og_fmt.with_columns(rtype=pl.lit("left")),
            df_right_og_fmt.with_columns(rtype=pl.lit("right")),
        ]
    )
    cens: dict[str, set[RowVals]] = {"left": set(), "right": set()}
    dupe_cens: dict[str, set[RowVals]] = {"left": set(), "right": set()}
    shared_cens: set[tuple[RowVals, RowVals]] = set()
    covered_cens: set[str] = set()

    # Haplotype information is unreliable so have to do row-by-row comparison.
    # Full outer join makes things very difficult.
    for _, df_grp in df_merged_og_fmt.group_by(["sample", "chr"]):
        df_grp = df_grp.filter(~pl.col("og_ctg").is_in(covered_cens))
        for row_1 in df_grp.iter_rows(named=True):
            row_1_vals = tuple(row_1.values())
            for row_2 in df_grp.iter_rows(named=True):
                row_2_vals = tuple(row_2.values())
                if (
                    row_1["og_ctg"] == row_2["og_ctg"]
                    or row_2["og_ctg"] in covered_cens
                ):
                    continue

                abs_bp_diff = abs(row_1["length"] - row_2["length"])
                # Get potential duplicates.
                # Very rare for AS-HOR array length to be exactly identical.
                if abs_bp_diff == 0 and row_1["rtype"] == row_2["rtype"]:
                    dupe_cens[row_1["rtype"]].add(row_1_vals)
                    dupe_cens[row_1["rtype"]].add(row_2_vals)

                if abs_bp_diff > bp_diff:
                    continue

                if row_1["og_ctg"] in covered_cens:
                    continue

                covered_cens.add(row_1["og_ctg"])
                covered_cens.add(row_2["og_ctg"])
                shared_cens.add((row_1_vals, row_2_vals))

            if row_1["og_ctg"] in covered_cens:
                continue

            cens[row_1["rtype"]].add(row_1_vals)
            covered_cens.add(row_1["og_ctg"])

    del covered_cens

    new_schema = df_left_og_fmt.schema
    new_schema["rtype"] = pl.String
    df_left = pl.DataFrame(list(cens["left"]), schema=new_schema, orient="row").select(
        **select_exprs()
    )
    df_right = pl.DataFrame(
        list(cens["right"]), schema=new_schema, orient="row"
    ).select(**select_exprs())

    shared_left, shared_right = zip(*list(shared_cens))
    df_shared = pl.concat(
        [
            pl.DataFrame(shared_left, schema=new_schema, orient="row").select(
                **select_exprs()
            ),
            pl.DataFrame(
                shared_right,
                schema=[f"{s}_right" for s in new_schema.keys()],
                orient="row",
            ).select(**select_exprs(side=Side.Right)),
        ],
        how="horizontal",
    )
    df_left_potential_dupes = (
        pl.DataFrame(list(dupe_cens["left"]), schema=new_schema, orient="row")
        .select(**select_exprs())
        .sort(by=["ctg", "length"])
    )
    df_right_potential_dupes = (
        pl.DataFrame(list(dupe_cens["right"]), schema=new_schema, orient="row")
        .select(**select_exprs())
        .sort(by=["ctg", "length"])
    )

    if not df_left_potential_dupes.is_empty():
        logger.warning(
            f"{df_left_potential_dupes.shape[0]} centromeres potentially duplicated in left file."
        )
    if not df_right_potential_dupes.is_empty():
        logger.warning(
            f"{df_right_potential_dupes.shape[0]} centromeres potentially duplicated in right file."
        )

    logger.info(f"{df_left.shape[0]} unique centromeres in left file.")
    logger.info(f"{df_right.shape[0]} unique centromeres in right file.")
    logger.info(f"{df_shared.shape[0]} centromeres shared by both files.")

    df_left.write_csv(outfile_left, include_header=False, separator="\t")
    df_right.write_csv(outfile_right, include_header=False, separator="\t")
    df_shared.write_csv(outfile_both, include_header=False, separator="\t")
    df_left_potential_dupes.write_csv(
        outfile_dupe_left, include_header=False, separator="\t"
    )
    df_right_potential_dupes.write_csv(
        outfile_dupe_right, include_header=False, separator="\t"
    )


def add_nonredundant_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "nonredundant",
        description="Get non-redundant list of centromeres based on HOR array length from two AS-HOR array length lists.",
    )
    ap.add_argument(
        "-l",
        "--infile_left",
        type=argparse.FileType("rb"),
        help=" ".join(
            [
                f"Centromere lengths. Expects columns: {IO_COLS}.",
                f"The '{IO_COLS[0]}' column must be able to be split by '_' into ['sample', 'chr', 'hap-ctg'].",
                "ex. 'HG01573_rc-chr1_haplotype1-0000024:121168122-126852171'",
            ]
        ),
        required=True,
    )
    ap.add_argument(
        "-r",
        "--infile_right",
        type=argparse.FileType("rb"),
        default=sys.stdin,
        help="Centromere lengths. Same format expected as --infile_left.",
        required=True,
    )
    ap.add_argument(
        "-ol",
        "--outfile_left",
        help="Unique centromeres from --infile_left.",
        type=str,
        default="uniq_left.tsv",
    )
    ap.add_argument(
        "-or",
        "--outfile_right",
        help="Unique centromeres from --infile_right.",
        type=str,
        default="uniq_right.tsv",
    )
    ap.add_argument(
        "-b",
        "--outfile_both",
        help="Centromeres shared by both files. Centromeres from --infile_left in first four columns and centromeres from --infile_right in last four columns.",
        type=str,
        default="both.tsv",
    )
    ap.add_argument(
        "-dl",
        "--duplicates_left",
        help="Potentially duplicated centromeres from --infile_left.",
        type=str,
        default="dupes_left.tsv",
    )
    ap.add_argument(
        "-dr",
        "--duplicates_right",
        help="Potentially duplicated centromeres from --infile_right.",
        type=str,
        default="dupes_right.tsv",
    )
    ap.add_argument(
        "-d",
        "--diff_bp",
        type=int,
        help=f"Difference in base pair length between two HOR arrays to be considered different. Defaults to {BP_DIFF} bp",
        default=BP_DIFF,
    )



DEF_MIN_BLK_HOR_UNITS = 2
DEF_MIN_ARR_HOR_UNITS = 10
# ~1 LINE element.
DEF_BP_MERGE_BLKS = 8000
# 1.5 monomers.
DEF_BP_MERGE_UNITS = 256
DEF_MIN_ARR_LEN = 30_000
DEF_MIN_ARR_PROP = 0.9

DEF_MERGE_RBLACKLIST = {"BSR", "GSAT", "SAR", "HSATII", "HSATI", "(CATTC)n", "(GAATG)n"}

DEF_INPUT_BED_COLS = ["chrom", "chrom_st", "chrom_end", "name", "score", "strand"]
DEF_INPUT_RM_COLS = [
    "contig",
    "start",
    "end",
    "type",
    "rClass",
]
DEF_INPUT_RM_COL_IDX = [4, 5, 6, 9, 10]

DEF_OUTPUT_BED_COLS = ["chrom", "chrom_st", "chrom_end", "name", "score", "prop"]
DEF_OUTPUT_BED_COLS_STRAND = [
    "chrom",
    "chrom_st",
    "chrom_end",
    "name",
    "score",
    "prop",
    "strand",
]



import re
import polars as pl
from typing import NamedTuple, Generator
from .constants import RGX_CHR


class RefCenContigs(NamedTuple):
    chr: str
    ref: str
    df: pl.DataFrame


def split_ref_rm_input_by_contig(
    df_ref: pl.DataFrame,
) -> Generator[tuple[str, RefCenContigs], None, None]:
    for ref, df_ref_grp in df_ref.group_by(["contig"]):
        ref = ref[0]
        mtch_ref_chr_name = re.search(RGX_CHR, ref)
        if not mtch_ref_chr_name:
            continue

        ref_chr_name = mtch_ref_chr_name.group()

        yield ref, RefCenContigs(ref_chr_name, ref, df_ref_grp)



import re
import polars as pl

from typing import Iterator, Optional

from .constants import RGX_CHR, REPEAT_SPLIT_LEN


def split_repeats(x: int, div: int) -> Iterator[int]:
    """
    Explodes/expands repeats per div.
    * ex. div = 1000
            * 2001 bp ALR = [1000 bp ALR, 1000 bp ALR, 1 bp ALR]
    """
    d, m = divmod(x, div)
    for d in range(d):
        yield div
    yield m


def expand_repeat_dst(
    df_ctg_grp: pl.DataFrame,
    *,
    repeat_filter: Optional[pl.Expr] = None,
    bp_repeat_split: int = REPEAT_SPLIT_LEN,
) -> pl.DataFrame:
    """
    Expand the repeat distance.

    ### Args
    `df_ctg_grp`
        RepeatMasker annotation dataframe of a single centromeric contig.
    `repeat_filter`
        Expression to filter and expand a subset of repeats.
    `bp_repeat_split`
        Number of base pairs to split the repeat by.

    ### Returns
    `pl.DataFrame` of expanded repeats.
    """
    return df_ctg_grp.with_columns(
        pl.when(repeat_filter if repeat_filter is not None else False)
        .then(
            pl.col("dst").map_elements(
                lambda x: list(split_repeats(x, bp_repeat_split)),
                return_dtype=pl.List(pl.Int64),
            )
        )
        .otherwise(pl.col("dst").cast(pl.List(pl.Int64)))
    ).explode("dst")


def get_contig_similarity_by_edit_dst(
    contigs: list[str],
    ref_contigs: list[str],
    edit_dst: list[int],
    orientation: list[str],
    *,
    dst_perc_thr: float,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    df_edit_distance_res = pl.DataFrame(
        {"contig": contigs, "ref": ref_contigs, "dst": edit_dst, "ort": orientation}
    ).with_columns(
        dst_perc=(pl.col("dst").rank() / pl.col("dst").count()).over("contig"),
    )

    # Filter results so only:
    # * Matches gt x percentile.
    # * Distances lt y percentile.
    dfs_filtered_edit_distance_res = []
    dfs_filtered_ort_same_chr_res = []

    edit_distance_thr_filter = pl.col("dst_perc") < dst_perc_thr
    edit_distance_lowest_dst_filter = pl.col("dst_perc") == pl.col("dst_perc").min()

    for contig, df_edit_distance_res_grp in df_edit_distance_res.group_by(["contig"]):
        mtch_chr_name = re.search(RGX_CHR, contig[0])
        if not mtch_chr_name:
            continue
        chr_name = mtch_chr_name.group()

        # Only look at same chr to determine default ort.
        df_edit_distance_res_same_chr_grp = df_edit_distance_res_grp.filter(
            pl.col("ref").str.contains(f"{chr_name}:")
        )

        df_filter_edit_distance_res_grp = df_edit_distance_res_grp.filter(
            edit_distance_thr_filter
        )
        df_filter_ort_res_same_chr_grp = df_edit_distance_res_same_chr_grp.filter(
            edit_distance_thr_filter
        )

        # If none found, default to highest number of matches.
        if df_filter_edit_distance_res_grp.is_empty():
            df_filter_edit_distance_res_grp = df_edit_distance_res_grp.filter(
                edit_distance_lowest_dst_filter
            )

        if df_filter_ort_res_same_chr_grp.is_empty():
            df_filter_ort_res_same_chr_grp = df_edit_distance_res_same_chr_grp.filter(
                edit_distance_lowest_dst_filter
            )

        dfs_filtered_edit_distance_res.append(df_filter_edit_distance_res_grp)
        dfs_filtered_ort_same_chr_res.append(df_filter_ort_res_same_chr_grp)

    df_filter_edit_distance_res: pl.DataFrame = pl.concat(
        dfs_filtered_edit_distance_res
    )
    df_filter_ort_same_chr_res: pl.DataFrame = pl.concat(dfs_filtered_ort_same_chr_res)

    df_filter_edit_distance_res = (
        df_filter_edit_distance_res
        # https://stackoverflow.com/a/74336952
        .with_columns(pl.col("dst").min().over("contig").alias("lowest_dst"))
        .filter(pl.col("dst") == pl.col("lowest_dst"))
        .select(["contig", "ref", "dst", "ort"])
    )
    # Get pair with lowest dst to get default ort.
    df_filter_ort_same_chr_res = (
        df_filter_ort_same_chr_res.with_columns(
            pl.col("dst").min().over("contig").alias("lowest_dst")
        )
        .filter(pl.col("dst") == pl.col("lowest_dst"))
        .select(["contig", "ort"])
        .rename({"ort": "ort_same_chr"})
    )

    return df_filter_edit_distance_res, df_filter_ort_same_chr_res



import polars as pl


def jaccard_index(a: set[str], b: set[str]) -> float:
    """
    Jaccard similarity index.
    * https://www.statisticshowto.com/jaccard-index/
    """
    return (len(a.intersection(b)) / len(a.union(b))) * 100.0


def get_contig_similarity_by_jaccard_index(
    contigs: list[str], ref_contigs: list[str], jaccard_index: list[float]
) -> pl.DataFrame:
    return (
        pl.LazyFrame(
            {"contig": contigs, "ref": ref_contigs, "similarity": jaccard_index}
        )
        .with_columns(
            pl.col("similarity").max().over("contig").alias("highest_similarity")
        )
        .filter(pl.col("similarity") == pl.col("highest_similarity"))
        .select(["contig", "ref", "similarity"])
        .collect()
    )



import polars as pl

from typing import TextIO
from censtats.length.constants import (
    DEF_INPUT_BED_COLS,
    DEF_INPUT_RM_COLS,
    DEF_INPUT_RM_COL_IDX,
)


def read_rm(infile: TextIO | str) -> pl.DataFrame:
    """
    Read a tab-delimited RepeatMasker output file and adjust relative coordinates to absolute coordinates.
    """
    return (
        pl.read_csv(
            infile,
            separator="\t",
            has_header=False,
            columns=DEF_INPUT_RM_COL_IDX,
            new_columns=DEF_INPUT_RM_COLS,
            truncate_ragged_lines=True,
        )
        .with_columns(
            ctg_name=pl.col("contig").str.extract(r"^(.*?):|^(.*?)$"),
            ctg_st=pl.col("contig").str.extract(r":(\d+)-").cast(pl.Int64).fill_null(0),
            ctg_end=pl.col("contig")
            .str.extract(r"-(\d+)$")
            .cast(pl.Int64)
            .fill_null(0),
        )
        # Adjust for contig coordinates if any.
        .with_columns(
            pl.col("start") + pl.col("ctg_st"), pl.col("end") + pl.col("ctg_st")
        )
    )


def read_stv(infile: TextIO | str) -> pl.DataFrame:
    """
    Read an HOR Stv bed.
    """
    return pl.read_csv(
        infile,
        separator="\t",
        columns=[0, 1, 2, 3, 4, 5],
        new_columns=DEF_INPUT_BED_COLS,
        has_header=False,
    )


def format_and_output_lengths(
    df: pl.DataFrame,
    output: TextIO | str,
    output_cols: list[str],
) -> None:
    if df.is_empty():
        raise ValueError("No live HOR data.")
    (
        df.with_columns(
            sort_idx=pl.col("chrom")
            .str.extract(r"chr([0-9XY]+)")
            .replace({"X": "23", "Y": "24"})
            .cast(pl.Int32)
        )
        .sort(by="sort_idx")
        .select(output_cols)
        .write_csv(output, include_header=False, separator="\t")
    )



import polars as pl

from .constants import EDGE_LEN, EDGE_PERC_ALR_THR, MAX_ALR_LEN_THR


def is_partial_centromere(
    df: pl.DataFrame,
    *,
    edge_len: int = EDGE_LEN,
    edge_perc_alr_thr: float = EDGE_PERC_ALR_THR,
    max_alr_len_thr: int = MAX_ALR_LEN_THR,
) -> bool:
    """
    Check if centromere is partially constructed based on ALR percentage at either ends of the contig.

    ALR/Alpha repeat content is used as a heuristic as we expect primarily monomeric repeats to be at the edges of the HOR array.

    ### Args
    `df`
        RepeatMasker output with a `dst` column for each repeat row.
    `edge_len`
        Edge len to check. Defaults to 100 kbp.
    `edge_perc_alr_thr`
        ALR percentage threshold needed to be considered incomplete. Defaults to 70%.
    `max_alr_len_thr`
        Length threshold of largest ALR needed to be considered complete. Defaults to 200 kbp.

    ### Returns
    Whether the centromere is partially constructed.
    """
    # Check if partial centromere based on ALR perc on ends.
    # Check N kbp from start and end of contig.
    ledge = (
        df.filter(pl.col("start") < edge_len)
        # Ensure that don't get more than edge length.
        .with_columns(
            end=pl.when(pl.col("end") > edge_len)
            .then(pl.lit(edge_len))
            .otherwise(pl.col("end"))
        )
        .with_columns(dst=pl.col("end") - pl.col("start"))
    )
    redge = df.filter(pl.col("start") > df[-1]["end"] - edge_len)
    try:
        ledge_perc_alr = (
            ledge.group_by("type")
            .agg(pl.col("dst").sum() / ledge["dst"].sum())
            .filter(pl.col("type") == "ALR/Alpha")
            .row(0)[1]
        )
    except pl.exceptions.OutOfBoundsError:
        ledge_perc_alr = 0.0 if not ledge.is_empty() else 100.0
    try:
        redge_perc_alr = (
            redge.group_by("type")
            .agg(pl.col("dst").sum() / redge["dst"].sum())
            .filter(pl.col("type") == "ALR/Alpha")
            .row(0)[1]
        )
    except pl.exceptions.OutOfBoundsError:
        redge_perc_alr = 0.0 if not redge.is_empty() else 100.0

    # Check if edges have ALR.
    are_edges_alr = (
        ledge_perc_alr > edge_perc_alr_thr or redge_perc_alr > edge_perc_alr_thr
    )
    # If they don't, check that the contig has at least one ALR that meets the minimum threshold len.
    if not are_edges_alr:
        max_alr_len = df.filter(pl.col("type") == "ALR/Alpha").get_column("dst").max()
        if not max_alr_len:
            return True
        return max_alr_len < max_alr_len_thr

    return are_edges_alr



from collections import deque
from typing import Callable, Iterable

from intervaltree import Interval


def fn_cmp_def(_itv_1: Interval, _itv_2: Interval) -> bool:
    return True


def fn_merge_itv_def(itv_1: Interval, itv_2: Interval) -> Interval:
    return Interval(begin=itv_1.begin, end=itv_2.end, data=None)


def merge_itvs(
    itvs: Iterable[Interval],
    dst: int = 1,
    fn_cmp: Callable[[Interval, Interval], bool] | None = None,
    fn_merge_itv: Callable[[Interval, Interval], Interval] | None = None,
) -> list[Interval]:
    if not fn_cmp:
        fn_cmp = fn_cmp_def
    if not fn_merge_itv:
        fn_merge_itv = fn_merge_itv_def

    final_itvs = []
    sorted_itvs = deque(sorted(itvs))
    while sorted_itvs:
        try:
            itv_1 = sorted_itvs.popleft()
        except IndexError:
            break
        try:
            itv_2 = sorted_itvs.popleft()
        except IndexError:
            final_itvs.append(itv_1)
            break
        dst_between = itv_2.begin - itv_1.end
        passes_cmp = fn_cmp(itv_1, itv_2)
        if dst_between <= dst and passes_cmp:
            sorted_itvs.appendleft(fn_merge_itv(itv_1, itv_2))
        else:
            final_itvs.append(itv_1)
            sorted_itvs.appendleft(itv_2)

    return final_itvs
